#-----------------------------------------------------------------------------------------------------
# BEGIN GPL LICENSE BLOCK
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# END GPL LICENCE BLOCK
#-----------------------------------------------------------------------------------------------------

bl_info = {
	"name": "BMax Tools",
	"author": "Ozzkar",
	"version": (0, 3, 0),
	"blender": (2, 83, 9),
	"location": "View3D > BMax",
	"description": "3ds Max/Maya inspired UI for Blender",
	"warning": "",
	"wiki_url": "",
	"tracker_url": "",
	"category": "3D View"}

#-----------------------------------------------------------------------------------------------------

import bpy
import bmesh
import math, mathutils
from mathutils import Vector, Matrix
from bpy.types import Panel, Menu, UIList, Operator, PropertyGroup
from bpy.props import IntProperty, BoolProperty, StringProperty, EnumProperty, FloatProperty, FloatVectorProperty

#-----------------------------------------------------------------------------------------------------
# Object Mode : Align
#-----------------------------------------------------------------------------------------------------

def _bmax_align_get_data(obj):
	# get AABB
	cld = []
	msh = obj.data
	if obj.type == 'MESH' and len(msh.vertices) > 0:
		for vert in msh.vertices:
			cld.append(obj.matrix_world @ vert.co)
	elif obj.type == 'CURVE' and len(msh.splines) > 0:
		for spn in msh.splines:
			for pts in spn.bezier_points:
				cld.append(obj.matrix_world @ pts.co)
	elif obj.type == 'SURFACE' and len(msh.splines) > 0:
		for spn in msh.splines:
			for pts in spn.points:
				cld.append(obj.matrix_world @ pts.co)
	elif obj.type == 'FONT' and len(msh.splines) > 0:
		for s in msh.splines:
			for pts in s.bezier_points:
				cld.append(obj.matrix_world @ pts.co)
	# get min/max/center/pivot data
	if len(cld) == 0 or obj.type not in {'MESH', 'CURVE', 'SURFACE', 'FONT'}:
		return obj.location.copy(), obj.location.copy(), obj.location.copy(), obj.location.copy()
	p_min = cld[0].copy()
	p_max = cld[0].copy()
	for v in cld:
		if p_min.x > v.x: p_min.x = v.x
		if p_min.y > v.y: p_min.y = v.y
		if p_min.z > v.z: p_min.z = v.z
		if p_max.x < v.x: p_max.x = v.x
		if p_max.y < v.y: p_max.y = v.y
		if p_max.z < v.z: p_max.z = v.z
	return p_min, (p_min + ((p_max - p_min) / 2)), obj.location.copy(), p_max

#-----------------------------------------------------------------------------------------------------

def _bmax_align_store_data(self, ctx):
	self.pos_curs = ctx.scene.cursor.location.copy()
	for obj in ctx.selected_objects:
		p_min, p_mid, p_piv, p_max = _bmax_align_get_data(obj)
		self.pos_list.append([p_min, p_mid, p_piv, p_max])
		self.rot_list.append(obj.rotation_euler.copy())
		self.scl_list.append(obj.scale.copy())

#-----------------------------------------------------------------------------------------------------

def _bmax_align_execute(self, ctx):
	a_box = _bmax_align_get_data(ctx.active_object)
	a_rot = ctx.active_object.rotation_euler
	a_scl = ctx.active_object.scale
	# set modes
	c_id = 0
	t_id = 0
	if self.c_mode == 'MID': c_id = 1
	if self.c_mode == 'PIV': c_id = 2
	if self.c_mode == 'MAX': c_id = 3
	if self.t_mode == 'MID': t_id = 1
	if self.t_mode == 'PIV': t_id = 2
	if self.t_mode == 'MAX': t_id = 3
	# check cursor to cursor
	if self.c_mode == '3DC' and self.t_mode == '3DC':
		return
	# check active object agains itself
	if len(ctx.selected_objects) == 1 and self.c_mode != '3DC' and self.t_mode != '3DC':
		return
	# set 3D cursor
	if self.c_mode == '3DC':
		if self.pos_x == True: ctx.scene.cursor.location.x = a_box[t_id].x
		if self.pos_y == True: ctx.scene.cursor.location.y = a_box[t_id].y
		if self.pos_z == True: ctx.scene.cursor.location.z = a_box[t_id].z
		return
	if self.t_mode == '3DC':
		scp = ctx.scene.cursor.location.copy()
		for i, obj in enumerate(ctx.selected_objects):
			if self.pos_x == True: obj.location.x = scp.x + (self.pos_list[i][2].x - self.pos_list[i][c_id].x)
			if self.pos_y == True: obj.location.y = scp.y + (self.pos_list[i][2].y - self.pos_list[i][c_id].y)
			if self.pos_z == True: obj.location.z = scp.z + (self.pos_list[i][2].z - self.pos_list[i][c_id].z)
		return
	# set selection
	for i, obj in enumerate(ctx.selected_objects):
		if obj != ctx.active_object:
			# position
			if self.pos_x == True: obj.location.x = a_box[t_id].x + (self.pos_list[i][2].x - self.pos_list[i][c_id].x)
			if self.pos_y == True: obj.location.y = a_box[t_id].y + (self.pos_list[i][2].y - self.pos_list[i][c_id].y)
			if self.pos_z == True: obj.location.z = a_box[t_id].z + (self.pos_list[i][2].z - self.pos_list[i][c_id].z)
			# rotation
			if self.rot_x == True: obj.rotation_euler.x = a_rot.x
			if self.rot_y == True: obj.rotation_euler.y = a_rot.y
			if self.rot_z == True: obj.rotation_euler.z = a_rot.z
			# scale
			if self.scl_x == True: obj.scale.x = a_scl.x
			if self.scl_y == True: obj.scale.y = a_scl.y
			if self.scl_z == True: obj.scale.z = a_scl.z

#-----------------------------------------------------------------------------------------------------

def _bmax_align_reset(self, ctx):
	ctx.scene.cursor.location.x = self.pos_curs.x
	ctx.scene.cursor.location.y = self.pos_curs.y
	ctx.scene.cursor.location.z = self.pos_curs.z
	for i, obj in enumerate(ctx.selected_objects):
		# position
		obj.location.x = self.pos_list[i][2].x
		obj.location.y = self.pos_list[i][2].y
		obj.location.z = self.pos_list[i][2].z
		# rotation
		obj.rotation_euler.x = self.rot_list[i].x
		obj.rotation_euler.y = self.rot_list[i].y
		obj.rotation_euler.z = self.rot_list[i].z
		# scale
		obj.scale.x = self.scl_list[i].x
		obj.scale.y = self.scl_list[i].y
		obj.scale.z = self.scl_list[i].z

	ctx.view_layer.update()

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_object_align(Operator):
	bl_idname = "bmax.align_tool"
	bl_label = "Align Selection"
	bl_description = "Align selection dialog box"
	bl_options = {'REGISTER', 'UNDO'}

	pos_curs = Vector((0, 0, 0))
	pos_list = []
	rot_list = []
	scl_list = []

	c_mode: EnumProperty(
		name = '', description = 'Selection alignment point', default = 'MIN',
		items = [
			('MIN', 'Minimum', 'Minimum'),
			('MID', 'Center', 'Center'),
			('PIV', 'Pivot', 'Pivot'),
			('MAX', 'Maximum', 'Maximum'),
			('3DC', '3D Cursor', '3D Cursor')]
		)
	t_mode: EnumProperty(
		name = '', description = 'Target alignment point', default = 'MIN',
		items = [
			('MIN', 'Minimum', 'Minimum'),
			('MID', 'Center', 'Center'),
			('PIV', 'Pivot', 'Pivot'),
			('MAX', 'Maximum', 'Maximum'),
			('3DC', '3D Cursor', '3D Cursor')]
		)
	pos_x: BoolProperty(default = False)
	pos_y: BoolProperty(default = False)
	pos_z: BoolProperty(default = False)
	rot_x: BoolProperty(default = False)
	rot_y: BoolProperty(default = False)
	rot_z: BoolProperty(default = False)
	scl_x: BoolProperty(default = False)
	scl_y: BoolProperty(default = False)
	scl_z: BoolProperty(default = False)

	@classmethod
	def poll(self, ctx):
		return len(ctx.selected_objects) > 0 and ctx.active_object != None and ctx.mode == 'OBJECT'

	def check(self, ctx):
		_bmax_align_reset(self, ctx)
		_bmax_align_execute(self, ctx)
		return True

	def draw(self, ctx):
		ui = self.layout
		b = ui.box()
		r = b.row()
		r.label(text="Align Position (World):")
		r = b.row()
		r.prop(self, "pos_x", text = "X Position")
		r.prop(self, "pos_y", text = "Y Position")
		r.prop(self, "pos_z", text = "Z Position")
		r = b.row()
		r.label(text="Current Object:")
		r.label(text="Taget Object:")
		r = b.row()
		r.prop(self, "c_mode")
		r.prop(self, "t_mode")
		b = ui.box()
		r = b.row()
		r.label(text="Align Orientation (Local):")
		r = b.row()
		r.prop(self, "rot_x", text = "X Axis")
		r.prop(self, "rot_y", text = "Y Axis")
		r.prop(self, "rot_z", text = "Z Axis")
		b = ui.box()
		r = b.row()
		r.label(text="Match Scale:")
		r = b.row()
		r.prop(self, "scl_x", text = "X Axis")
		r.prop(self, "scl_y", text = "Y Axis")
		r.prop(self, "scl_z", text = "Z Axis")

	def execute(self, ctx):
		return {'FINISHED'}

	def cancel(self, ctx):
		_bmax_align_reset(self, ctx)

	def invoke(self, ctx, evt):
		self.pos_curs = Vector((0, 0, 0))
		self.pos_list = []
		self.rot_list = []
		self.scl_list = []
		_bmax_align_store_data(self, ctx)
		_bmax_align_execute(self, ctx)
		ctx.window_manager.invoke_props_dialog(self)
		return {'RUNNING_MODAL'}

#-----------------------------------------------------------------------------------------------------
# Object Mode : Clone
#-----------------------------------------------------------------------------------------------------

def _bmax_clone_reset(self, ctx):
	# clear selection
	bpy.ops.object.select_all(action='DESELECT')
	# delete temporary
	for obj in self.new_list:
		obj.select_set(True)
		ctx.view_layer.objects.active = obj
		bpy.ops.object.delete(use_global=False, confirm=False)
	# reselect initial object
	self.new_list = []
	self.obj_orig.select_set(True)
	ctx.view_layer.objects.active = self.obj_orig

#-----------------------------------------------------------------------------------------------------

def _bmax_clone_execute(self, ctx):
	# process parameters
	if self.new_size < 1:
		self.new_size = 1;
	pos_x = self.new_move[0]
	pos_y = self.new_move[1]
	pos_z = self.new_move[2]
	scl_x = self.new_scls[0]
	scl_y = self.new_scls[1]
	scl_z = self.new_scls[2]

	if self.new_xpos == False:
		pos_x = 0
	if self.new_ypos == False:
		pos_y = 0
	if self.new_zpos == False:
		pos_z = 0

	off_value = (pos_x, pos_y, pos_z)

	if self.new_xscl == False:
		scl_x = 1
	if self.new_yscl == False:
		scl_y = 1
	if self.new_zscl == False:
		scl_z = 1

	off_scale = (scl_x, scl_y, scl_z)

	# create clone preview
	for i in range(self.new_size):
		bpy.ops.object.duplicate(linked = not self.new_copy, mode = 'INIT')
		# scale
		bpy.ops.transform.resize(value = off_scale)
		# rotate
		if self.new_xrot == True:
			ctx.active_object.rotation_euler[0] += math.radians(self.new_turn[0])
		if self.new_yrot == True:
			ctx.active_object.rotation_euler[1] += math.radians(self.new_turn[1])
		if self.new_zrot == True:
			ctx.active_object.rotation_euler[2] += math.radians(self.new_turn[2])
		# translate
		bpy.ops.transform.translate(value = off_value, orient_type = self.pos_mode)

		if self.new_name != self.obj_orig.name and i == 0:
			ctx.active_object.name = self.new_name
		self.new_list.append(ctx.active_object)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_object_clone(Operator):
	bl_idname = "bmax.clone_tool"
	bl_label = "Clone Options"
	bl_description = "Clone object dialog box"
	bl_options = {'UNDO'}

	new_name: StringProperty(name= 'Autoname prefix', default = "Default")
	new_size: IntProperty(default = 1, soft_min = 1)
	new_copy: BoolProperty(default = True)
	# translation offset
	pos_mode: EnumProperty(
		name = 'Coord. System', 
		description = 'Coordinate System', 
		default = 'GLOBAL',
		items = [('GLOBAL', 'Global', 'Global'), ('LOCAL', 'Local', 'Local')]
	)
	new_xpos: BoolProperty(default = False)
	new_ypos: BoolProperty(default = False)
	new_zpos: BoolProperty(default = False)
	new_move: FloatVectorProperty(default = (0, 0, 0), subtype='XYZ')
	# rotation offset
	new_xrot: BoolProperty(default = False)
	new_yrot: BoolProperty(default = False)
	new_zrot: BoolProperty(default = False)
	new_turn: FloatVectorProperty(default = (0, 0, 0), subtype='XYZ')
	# scale offset
	new_xscl: BoolProperty(default = False)
	new_yscl: BoolProperty(default = False)
	new_zscl: BoolProperty(default = False)
	new_scls: FloatVectorProperty(default = (1, 1, 1), subtype='XYZ')
	# internal
	new_list = []
	obj_orig = 0

	@classmethod
	def poll(self, ctx):
		return len(ctx.selected_objects) == 1 and ctx.active_object != None and ctx.mode == 'OBJECT'

	def check(self, ctx):
		_bmax_clone_reset(self, ctx)
		_bmax_clone_execute(self, ctx)
		return True

	def draw(self, ctx):
		bx = self.layout.box()
		bx.row().prop(self, "new_name", text = "Name")
		bx.row().prop(self, "new_copy", text = "Copy")
		bx.row().prop(self, "new_size", text = "Count")
		s = bx.row().split(factor=0.3)
		s.column().label(text="Coord. System:")
		s.column().prop(self, "pos_mode", text="")
		# translation
		bx = self.layout.box()
		bx.row().label(text="Translation Offset:")
		s = bx.row().split(factor=0.75, align=True)
		s.column().prop(self, "new_move", text="")
		c = s.column(align=True)
		c.prop(self, "new_xpos", toggle = 1, text = "Apply")
		c.prop(self, "new_ypos", toggle = 1, text = "Apply")
		c.prop(self, "new_zpos", toggle = 1, text = "Apply")
		# rotation
		bx = self.layout.box()
		bx.row().label(text="Rotation Offset:")
		s = bx.row().split(factor=0.75, align=True)
		s.column().prop(self, "new_turn", text="")
		c = s.column(align=True)
		c.prop(self, "new_xrot", toggle = 1, text = "Apply")
		c.prop(self, "new_yrot", toggle = 1, text = "Apply")
		c.prop(self, "new_zrot", toggle = 1, text = "Apply")
		# scale
		bx = self.layout.box()
		bx.row().label(text="Scale Offset:")
		s = bx.row().split(factor=0.75, align=True)
		s.column().prop(self, "new_scls", text="")
		c = s.column(align=True)
		c.prop(self, "new_xscl", toggle = 1, text = "Apply")
		c.prop(self, "new_yscl", toggle = 1, text = "Apply")
		c.prop(self, "new_zscl", toggle = 1, text = "Apply")

	def execute(self, ctx):
		return {'FINISHED'}

	def cancel(self, ctx):
		_bmax_clone_reset(self, ctx)

	def invoke(self, ctx, evt):
		self.new_name = ctx.active_object.name
		self.obj_orig = ctx.active_object
		self.new_list = []
		_bmax_clone_execute(self, ctx)
		ctx.window_manager.invoke_props_dialog(self)
		return {'RUNNING_MODAL'}

#-----------------------------------------------------------------------------------------------------
# Object Mode : Mirror
#-----------------------------------------------------------------------------------------------------

def _bmax_mirror_object_execute(self, ctx):
	ctx.active_object.matrix_world = self.old_xfrm.copy()
	# set mode
	if self.c_mode == 'COPY':
		if self.obj_copy == None:
			bpy.ops.object.duplicate(linked = False, mode = 'INIT')
			self.obj_copy = ctx.active_object
	elif self.c_mode == 'INST':
		if self.obj_copy == None:
			bpy.ops.object.duplicate(linked = True, mode = 'INIT')
			self.obj_copy = ctx.active_object
	else:
		if self.obj_copy != None:
			bpy.ops.object.delete(use_global = False)
			ctx.view_layer.objects.active = self.obj_orig
			self.obj_orig.select_set(True)
			self.obj_copy = None
	# make transforms
	c_ax = (False, False, False)
	offs = (self.v_offs, self.v_offs, self.v_offs)
	if	 self.t_mode == 'X':
		c_ax = (True, False, False)
		offs = (self.v_offs, 0, 0)
	elif self.t_mode == 'Y':
		c_ax = (False, True, False)
		offs = (0, self.v_offs, 0)
	elif self.t_mode == 'Z':
		c_ax = (False, False, True)
		offs = (0, 0, self.v_offs)
	elif self.t_mode == 'XY':
		c_ax = (True, True, False)
		offs = (self.v_offs, self.v_offs, 0)
	elif self.t_mode == 'YZ':
		c_ax = (False, True, True)
		offs = (0, self.v_offs, self.v_offs)
	elif self.t_mode == 'ZX':
		c_ax = (True, False, True)
		offs = (self.v_offs, 0, self.v_offs)
	if self.t_mode != 'N':
		# mirror
		bpy.ops.transform.mirror(constraint_axis=c_ax, orient_type=self.dlg_csys.type)
		#offset
		bpy.ops.transform.translate(value=offs, orient_type=self.dlg_csys.type)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_object_mirror(Operator):
	bl_idname = "bmax.mirror_tool"
	bl_label = "Mirror"
	bl_description = "Mirror object dialog box"
	bl_options = {'UNDO'}

	old_xfrm = None
	obj_orig = None
	obj_copy = None
	dlg_csys = None

	v_offs: FloatProperty(name = 'Mirror offset', default = 0)
	t_mode: EnumProperty(
		name = 'Axis', 
		description = 'Mirror axis', 
		default = 'N',
		items = [
			('N', 'None', 'None'),
			('X', 'X', 'X'),
			('Y', 'Y', 'Y'),
			('Z', 'Z', 'Z'),
			('XY', 'XY', 'XY'),
			('YZ', 'YZ', 'YZ'),
			('ZX', 'ZX', 'ZX')]
		)
	c_mode: EnumProperty(
		name = 'Clone Selection', description = 'Clone mode', default = 'ORIG',
		items = [('ORIG', 'No Clone', 'No Clone'),('COPY', 'Copy', 'Copy'),('INST', 'Instance', 'Instance')]
		)

	@classmethod
	def poll(self, ctx):
		return len(ctx.selected_objects) == 1 and ctx.active_object != None and ctx.mode == 'OBJECT'

	def check(self, ctx):
		_bmax_mirror_object_execute(self, ctx)
		return True

	def cancel(self, ctx):
		if self.obj_copy != None:
			bpy.ops.object.delete(use_global = False)
			ctx.view_layer.objects.active = self.obj_orig
		self.obj_orig.select_set(True)
		self.obj_orig.matrix_world = self.old_xfrm.copy()

	def draw(self, ctx):
		ui = self.layout
		b = ui.box()
		s = b.split(factor=0.35)
		c1 = s.column()
		c2 = s.column()
		c1.row().label(text="Coord. System:")
		c1.row().label(text="Axis:")
		c1.row().label(text="Offset:")
		c1.row().label(text="Clone Selection:")
		c2.row().prop(self.dlg_csys, "type", text = "")
		c2.row().prop(self, "t_mode", text = "")
		c2.row().prop(self, "v_offs", text = "")
		c2.row().prop(self, "c_mode", text = "")

	def execute(self, ctx):
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		self.obj_copy = None
		self.obj_orig = ctx.active_object
		self.old_xfrm = self.obj_orig.matrix_world.copy()
		self.dlg_csys = ctx.window.scene.transform_orientation_slots[0]
		_bmax_mirror_object_execute(self, ctx)
		ctx.window_manager.invoke_props_dialog(self)
		return {'RUNNING_MODAL'}

#-----------------------------------------------------------------------------------------------------
# Object Mode : Align Special
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_object_align_special(Menu):
	bl_label = "Align Special"
	bl_description = "Pivot and 3D Cursor manipulation operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("object.origin_set", text = "Object To Pivot").type='GEOMETRY_ORIGIN'
		ui.operator("object.origin_set", text = "Pivot To Object").type='ORIGIN_GEOMETRY'
		ui.operator("object.origin_set", text = "Pivot To Cursor").type='ORIGIN_CURSOR'
		ui.operator("object.origin_set", text = "Pivot To COM").type='ORIGIN_CENTER_OF_MASS'
		ui.separator()
		ui.operator("view3d.snap_cursor_to_center", text = "Cursor To Center")
		ui.operator("view3d.snap_cursor_to_active", text = "Cursor To Active")
		ui.operator("view3d.snap_cursor_to_selected", text = "Cursor To Selection")
		ui.operator("view3d.snap_cursor_to_grid", text = "Cursor To Grid")

#-----------------------------------------------------------------------------------------------------
# Object Mode : Unhide All / Hide (Un)selected
#-----------------------------------------------------------------------------------------------------

# Unhide All
class BMAX_OT_object_unhide_all(Operator):
	bl_idname = "bmax.unhide_all"
	bl_label = "Unhide All"
	bl_description = "Unhide all hidden objects"

	def execute(self, ctx):
		for obj in ctx.scene.objects:
			if obj.hide_get() == True:
				obj.hide_set(False)
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		return self.execute(ctx)

# Hide Selected
class BMAX_OT_object_hide_selected(Operator):
	bl_idname = "bmax.hide_selected"
	bl_label = "Hide Selected"
	bl_description = "Hide all selected objects"

	def execute(self, ctx):
		for obj in ctx.selected_objects:
			obj.hide_set(True)
		for obj in ctx.scene.objects:
			if obj.select_get() == True:
				obj.select_set(False)
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		return self.execute(ctx)

# Hide Unselected
class BMAX_OT_object_hide_unselected(Operator):
	bl_idname = "bmax.hide_unselected"
	bl_label = "Hide Unselected"
	bl_description = "Hide all unselected objects"

	def execute(self, ctx):
		for obj in ctx.scene.objects:
			if obj.select_get() == False:
				obj.hide_set(True)
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

# Sub-Menu for last 3 operators
class BMAX_MT_object_hide(Menu):
	bl_label = "Hide/Unhide"
	bl_description = "Hide/Unhide operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("bmax.hide_selected")
		ui.operator("bmax.hide_unselected")
		ui.separator()
		ui.operator("bmax.unhide_all")

#-----------------------------------------------------------------------------------------------------
# Object Mode : Unfreeze All / Freeze (Un)selected
#-----------------------------------------------------------------------------------------------------

# Unfreeze All
class BMAX_OT_object_unfreeze_all(Operator):
	bl_idname = "bmax.unfreeze_all"
	bl_label = "Unfreeze All"
	bl_description = "Unfreeze all hidden objects"

	def execute(self, ctx):
		for obj in ctx.scene.objects:
			if obj.hide_select == True:
				obj.hide_select = False
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		return self.execute(ctx)

# Freeze Selected
class BMAX_OT_object_freeze_selected(Operator):
	bl_idname = "bmax.freeze_selected"
	bl_label = "Freeze Selected"
	bl_description = "Freeze all selected objects"

	def execute(self, ctx):
		for obj in ctx.selected_objects:
			obj.hide_select = True
		for obj in ctx.scene.objects:
			if obj.select_get() == True:
				obj.select_set(False)
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		return self.execute(ctx)

# Freeze Unselected
class BMAX_OT_object_freeze_unselected(Operator):
	bl_idname = "bmax.freeze_unselected"
	bl_label = "Freeze Unselected"
	bl_description = "Freeze all unselected objects"

	def execute(self, ctx):
		for obj in ctx.scene.objects:
			if obj.select_get() == False:
				obj.hide_select = True
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

# Sub-Menu for last 3 operators
class BMAX_MT_object_freeze(Menu):
	bl_label = "Freeze/Unfreeze"
	bl_description = "Freeze/Unfreeze operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("bmax.freeze_selected")
		ui.operator("bmax.freeze_unselected")
		ui.separator()
		ui.operator("bmax.unfreeze_all")

#-----------------------------------------------------------------------------------------------------
# Object Mode : Link
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_object_link(Menu):
	bl_label = "Link"
	bl_description = "Hierarchy manipulation operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("object.parent_set", text="Link", icon = "LINKED")
		ui.operator("object.parent_clear", text="Unlink", icon = "UNLINKED")
		ui.operator("object.parent_clear", text="Unlink Inverse").type = 'CLEAR_INVERSE'
		ui.operator("object.parent_clear", text="Unlink Keep XForm").type = 'CLEAR_KEEP_TRANSFORM'

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Align Special
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_make_planar_n(Operator):
	bl_idname = "bmax.make_planar_n"
	bl_label = ""
	bl_options = {'REGISTER', 'UNDO'}
	
	def invoke(self, ctx, evt):
		bpy.ops.transform.resize(value = (1,1,0), constraint_axis = (False, False, True), orient_type = 'NORMAL')
		return {'FINISHED'}

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_make_planar_x(Operator):
	bl_idname = "bmax.make_planar_x"
	bl_label = ""
	bl_options = {'REGISTER', 'UNDO'}
	
	def invoke(self, ctx, evt):
		bpy.ops.transform.resize(value = (0,1,1), constraint_axis = (True, False, False), orient_type = 'LOCAL')
		return {'FINISHED'}

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_make_planar_y(Operator):
	bl_idname = "bmax.make_planar_y"
	bl_label = ""
	bl_options = {'REGISTER', 'UNDO'}
	
	def invoke(self, ctx, evt):
		bpy.ops.transform.resize(value = (1,0,1), constraint_axis = (False, True, False), orient_type = 'LOCAL')
		return {'FINISHED'}

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_make_planar_z(Operator):
	bl_idname = "bmax.make_planar_z"
	bl_label = ""
	bl_options = {'REGISTER', 'UNDO'}
	
	def invoke(self, ctx, evt):
		bpy.ops.transform.resize(value = (1,1,0), constraint_axis = (False, False, True), orient_type = 'LOCAL')
		return {'FINISHED'}

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_view_align_x(Operator):
	bl_idname = "bmax.view_align_x"
	bl_label = ""
	bl_options = {'REGISTER', 'UNDO'}
	
	def invoke(self, ctx, evt):
		bpy.ops.transform.resize(value = (0,1,1), constraint_axis = (True, False, False), orient_type = 'VIEW')
		return {'FINISHED'}

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_view_align_y(Operator):
	bl_idname = "bmax.view_align_y"
	bl_label = ""
	bl_options = {'REGISTER', 'UNDO'}
	
	def invoke(self, ctx, evt):
		bpy.ops.transform.resize(value = (1,0,1), constraint_axis = (False, True, False), orient_type = 'VIEW')
		return {'FINISHED'}

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_view_align_z(Operator):
	bl_idname = "bmax.view_align_z"
	bl_label = ""
	bl_options = {'REGISTER', 'UNDO'}
	
	def invoke(self, ctx, evt):
		bpy.ops.transform.resize(value = (1,1,0), constraint_axis = (False, False, True), orient_type = 'VIEW')
		return {'FINISHED'}

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_mesh_align_special(Menu):
	bl_label = "Align Special"
	bl_description = "Align selection to specified plane/point"
	def draw(self, ctx):
		ui = self.layout
		# average normal
		op = ui.operator("bmax.make_planar_n", text = "Make Planar", icon = 'EMPTY_ARROWS')
		ui.separator()
		# local X
		op = ui.operator("bmax.make_planar_x", text = "Make Planar - X", icon = 'AXIS_SIDE')
		# local Y
		op = ui.operator("bmax.make_planar_y", text = "Make Planar - Y", icon = 'AXIS_FRONT')
		# local Z
		op = ui.operator("bmax.make_planar_z", text = "Make Planar - Z", icon = 'AXIS_TOP')
		ui.separator()
		# screen - X
		op = ui.operator("bmax.view_align_x", text = "View Align - X", icon = 'EMPTY_ARROWS')
		# screen - Z
		op = ui.operator("bmax.view_align_y", text = "View Align - Y", icon = 'EMPTY_ARROWS')
		# screen - Y
		op = ui.operator("bmax.view_align_z", text = "View Align - Z", icon = 'EMPTY_ARROWS')
		ui.separator()
		
		# Blender native
		ui.operator("view3d.snap_selected_to_cursor", text = "Selection To Cursor")
		ui.operator("view3d.snap_selected_to_grid", text = "Selection To Grid")
		ui.separator()
		ui.operator("view3d.snap_cursor_to_center", text = "Cursor To Center")
		ui.operator("view3d.snap_cursor_to_active", text = "Cursor To Active")
		ui.operator("view3d.snap_cursor_to_selected", text = "Cursor To Selection")
		ui.operator("view3d.snap_cursor_to_grid", text = "Cursor To Grid")

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Sub-Object Selection
#-----------------------------------------------------------------------------------------------------

# edit vertices
class BMAX_OT_mesh_sub_object_verts(Operator):
	bl_idname = "bmax.sub_object_verts"
	bl_label = "Edit Vertices"
	bl_description = "Select vertex sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.mesh_select_mode = (True, False, False)
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

# edit edges
class BMAX_OT_mesh_sub_object_edges(Operator):
	bl_idname = "bmax.sub_object_edges"
	bl_label = "Edit Edges"
	bl_description = "Select edge sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.mesh_select_mode = (False, True, False)
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

# edit faces
class BMAX_OT_mesh_sub_object_faces(Operator):
	bl_idname = "bmax.sub_object_faces"
	bl_label = "Edit Faces"
	bl_description = "Select face sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.mesh_select_mode = (False, False, True)
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

# edit edges+verts
class BMAX_OT_mesh_sub_object_edge_verts(Operator):
	bl_idname = "bmax.sub_object_edge_verts"
	bl_label = "Edit Edges/Vertices"
	bl_description = "Select edge/vertex sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.mesh_select_mode = (True, True, False)
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

# Sub-Menu for last 4 operators
class BMAX_MT_mesh_sub_object(Menu):
	bl_label = "Sub-Object"
	bl_description = "Select sub-object level"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("bmax.sub_object_verts", icon = 'VERTEXSEL')
		ui.operator("bmax.sub_object_edges", icon = 'EDGESEL')
		ui.operator("bmax.sub_object_faces", icon = 'FACESEL')
		ui.separator()
		ui.operator("bmax.sub_object_edge_verts", icon = 'SNAP_VERTEX')

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_mesh_select_special(Menu):
	bl_label = "Select Special"
	bl_description = "Select special sub-set"
	def draw(self, ctx):
		ui = self.layout
		if ctx.tool_settings.mesh_select_mode[1] == True:
			ui.operator("mesh.loop_multi_select", icon="UV_VERTEXSEL", text="Edge Loops").ring=False
			ui.operator("mesh.loop_multi_select", icon="UV_EDGESEL", text="Edge Rings").ring=True
		ui.operator("mesh.select_linked", icon="UV_ISLANDSEL", text="Select Element")

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Show/Hide Selection
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_mesh_show_hide_element(Menu):
	bl_label = "Hide/Unhide"
	bl_description = "Hide/Unhide operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("mesh.hide", text = "Hide Selected").unselected = False
		ui.operator("mesh.hide", text = "Hide Unselected").unselected = True
		ui.separator()
		ui.operator("mesh.reveal", text = "Unhide All")

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Clone
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_clone(Operator):
	bl_idname = "bmax.clone_mesh"
	bl_label = "Clone To Element"
	bl_description = "Clone selection into new element"
	def execute(self, ctx):
		bpy.ops.mesh.duplicate_move(MESH_OT_duplicate={"mode":1}, TRANSFORM_OT_translate={"value":(0, 0, 0)})
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_mesh_clone(Menu):
	bl_label = "Clone/Detach"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("bmax.clone_mesh", icon="MOD_BOOLEAN")
		ui.separator()
		ui.operator("mesh.split", text="Detach To Element", icon="MOD_BUILD")
		ui.operator("mesh.separate", text="Detach To Object", icon="MOD_BUILD")

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Create
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_mesh_create(Menu):
	bl_label = "Create"
	bl_description = "Create geometry operators"
	def draw(self, ctx):
		ui = self.layout
		mode = ctx.tool_settings.mesh_select_mode
		ui.operator("mesh.edge_face_add", text = "Face", icon = 'MESH_PLANE')
		ui.operator("mesh.fill", text = "Fill", icon = 'OUTLINER_DATA_MESH')
		ui.operator("mesh.fill_grid", text = "Grid Fill", icon = 'MESH_GRID')
		ui.operator("mesh.beautify_fill", text = "Beautify Fill", icon = 'MESH_ICOSPHERE')
		if mode[1] == True:
			ui.separator()
			ui.operator("mesh.bridge_edge_loops", text = "Bridge", icon = 'OUTLINER_DATA_LATTICE')

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Slice
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_mesh_slice(Menu):
	bl_label = "Slice"
	bl_description = "Slice geometry operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("mesh.knife_tool", text = "Cut", icon = "OUTLINER_DATA_CURVE")
		ui.separator()
		ui.operator("mesh.subdivide", text = "Subdivide")
		ui.operator("mesh.loopcut_slide", text = "Slice Plane")
		ui.operator("mesh.offset_edge_loops_slide", text="Offset Edge")
		ui.operator("mesh.bisect", text = "Bisect")
		ui.separator()
		ui.operator("mesh.rip", text = "Break Vertex", icon = "MOD_EDGESPLIT")
		ui.operator("mesh.edge_split", text = "Split Edge", icon = "MOD_EDGESPLIT")

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Mesh tools
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_turn_quad_diagonal(Operator):
	bl_idname = "bmax.turn_quad_diagonal"
	bl_label = "Turn Edge"
	bl_description = "Turns quad's hidden diagonal edge"

	@classmethod
	def poll(self, ctx):
		return ctx.object.data.total_face_sel == 1

	def execute(self, ctx):
		bpy.ops.mesh.quads_convert_to_tris(quad_method='FIXED_ALTERNATE')
		bpy.ops.mesh.tris_convert_to_quads()
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_mesh_meshtools(Menu):
	bl_label = "Mesh Tools"
	bl_description = "Selection of mesh tools"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("bmax.turn_quad_diagonal", icon="UV_SYNC_SELECT")
		ui.operator("mesh.edge_rotate", text="Rotate Edge CW", icon="UV_SYNC_SELECT").use_ccw=False
		ui.operator("mesh.edge_rotate", text="Rotate Edge CCW", icon="UV_SYNC_SELECT").use_ccw=True
		ui.separator()
		ui.operator("transform.shrink_fatten", text = "Shrink/Fatten", icon="MESH_MONKEY")
		ui.operator("transform.tosphere", text = "Spherify", icon="MESH_ICOSPHERE")
		ui.operator("transform.shear", text = "Shear")
		ui.operator("transform.push_pull", text = "Push/Pull", icon="MOD_MESHDEFORM")
		ui.operator("mesh.inset", text="Inset", icon="MOD_MESHDEFORM")
		ui.operator("mesh.vertices_smooth", text="Smooth", icon="MOD_SMOOTH")
		ui.operator("mesh.bevel", text="Bevel", icon="MOD_BEVEL")
		ui.separator()
		ui.operator("mesh.vert_connect", text="Connect")
		ui.operator("transform.edge_slide", text="Slide Edge")
		ui.operator("transform.vert_slide", text="Slide Vertex")
		ui.separator()
		ui.operator("mesh.remove_doubles", text="Weld", icon="MOD_DISPLACE")
		ui.operator("mesh.merge", text="Merge", icon="STICKY_UVS_DISABLE")

#-----------------------------------------------------------------------------------------------------
# Edit Mode : Target Weld
#-----------------------------------------------------------------------------------------------------

def _bmax_weld_check(ctx):
	b_snap = ctx.tool_settings.use_snap
	b_elem = ctx.tool_settings.snap_elements
	b_targ = ctx.tool_settings.snap_target
	b_auto = ctx.tool_settings.use_mesh_automerge
	b_self = ctx.tool_settings.use_snap_self
	return b_elem == {'VERTEX'} and b_targ == 'CLOSEST' and b_auto == True and b_snap == True and b_self == True

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_target_weld_on(Operator):
	bl_idname = "bmax.target_weld_on"
	bl_label = "Target Weld"
	bl_description = "Switch target weld on"
	def execute(self, ctx):
		ctx.tool_settings.snap_elements = {'VERTEX'}
		ctx.tool_settings.snap_target = 'CLOSEST'
		ctx.tool_settings.use_mesh_automerge = True
		ctx.tool_settings.use_snap_self = True
		ctx.tool_settings.use_snap = True
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_mesh_target_weld_off(Operator):
	bl_idname = "bmax.target_weld_off"
	bl_label = "Target Weld"
	bl_description = "Switch target weld off"
	def execute(self, ctx):
		ctx.tool_settings.snap_elements = {'INCREMENT'}
		ctx.tool_settings.use_mesh_automerge = False
		ctx.tool_settings.use_snap = False
		return {'FINISHED'}
	def invoke(self, ctx, event):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------
# Edit Mesh Mode : Repair
#-----------------------------------------------------------------------------------------------------

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_check_valid_faces(bpy.types.Operator):
	bl_idname = "bmax.check_valid_faces"
	bl_label = "Find N-Gons"
	bl_description = "Find polygons that are not triangles or quads"
	
	# face checker (find n-gons, etc.)
	def _check_valid_faces(obj):
		sz = 0
		me = obj.data
		bm = bmesh.from_edit_mesh(me)
		for f in bm.faces:
			if len(f.verts) != 3 and len(f.verts) != 4:
				f.select = True
				sz += 1
		bm.free()
		return sz

	@classmethod
	def poll(self, ctx):
		b1 = len(ctx.selected_objects) == 1
		b2 = ctx.active_object != None
		b3 = False
		if b2:
			b3 = ctx.active_object.type == 'MESH'
		return b1 and b2 and b3

	def execute(self, ctx):
		bpy.ops.object.mode_set(mode='EDIT')
		bpy.ops.mesh.select_all(action='DESELECT')
		sz = _check_valid_faces(ctx.active_object)
		bpy.ops.object.mode_set(mode='OBJECT')
		if sz == 0:
			self.report({'INFO'}, "No invalid faces")
		else:
			bpy.ops.object.mode_set(mode='EDIT')
			self.report({'ERROR'}, "Found " + str(sz) + " invalid faces")
		return {'FINISHED'}
		
	def invoke(self, ctx, evt):
		return self.execute(ctx)
		
#-----------------------------------------------------------------------------------------------------
# remove isolated geometry operator
class BMAX_OT_mesh_remove_isolated_geometry(Operator):
	bl_idname = "bmax.remove_isolated_geometry"
	bl_label = "Remove Isolated Geometry"
	bl_description = "Remove isolated geometry (vertices, edges, or faces)"
	def execute(self, ctx):
		mode = ctx.tool_settings.mesh_select_mode
		bpy.ops.mesh.select_loose()
		if mode[0]:
			bpy.ops.mesh.delete(type='VERT')
		if mode[1]:
			bpy.ops.mesh.delete(type='EDGE')
		if mode[2]:
			bpy.ops.mesh.delete(type='FACE')
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------
# repair sub-menu
class BMAX_MT_mesh_repair(Menu):
	bl_label = "Repair Geometry"
	bl_description = "Mesh repair operators"
	def draw(self, ctx):
		ui = self.layout
		labl = "Remove Isolated Geometry"
		mode = ctx.tool_settings.mesh_select_mode
		if mode[0] and not mode [1] and not mode[2]:
			labl = "Remove Isolated Vertices"
		elif mode[1] and not mode [0] and not mode[2]:
			labl = "Remove Isolated Edges"
		elif mode[2] and not mode [0] and not mode[1]:
			labl = "Remove Isolated Faces"
		ui.operator("bmax.remove_isolated_geometry", text = labl)
		ui.operator("mesh.flip_normals", text = "Flip Normals")
		ui.separator()
		ui.operator("bmax.check_valid_faces", text = "Select N-Gons")

#-----------------------------------------------------------------------------------------------------
# Edit Armature Mode : Create
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_armature_create_primitive(Menu):
	bl_label = "Create"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("armature.bone_primitive_add", text="Single Bone", icon='BONE_DATA')

#-----------------------------------------------------------------------------------------------------
# Edit Armature Mode : Align Special
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_armature_align_special(Menu):
	bl_label = "Align Special"
	bl_description = "Align selection to specified point"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("view3d.snap_selected_to_cursor", text = "Selection To Cursor")
		ui.operator("view3d.snap_selected_to_grid", text = "Selection To Grid")
		ui.separator()
		ui.operator("view3d.snap_cursor_to_center", text = "Cursor To Center")
		ui.operator("view3d.snap_cursor_to_active", text = "Cursor To Active")
		ui.operator("view3d.snap_cursor_to_selected", text = "Cursor To Selection")
		ui.operator("view3d.snap_cursor_to_grid", text = "Cursor To Grid")

#-----------------------------------------------------------------------------------------------------
# Edit Armature Mode : Bone Tools
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_armature_bone_tools(Menu):
	bl_label = "Bone Tools"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("armature.subdivide", text="Refine", icon="GROUP_BONE")
		ui.operator("armature.dissolve", text="Merge", icon="BONE_DATA")
		ui.operator("armature.fill", text="Bridge")
		ui.separator()
		ui.operator("armature.calculate_roll", text="Set Roll", icon="TRACKING_FORWARDS")
		ui.operator("transform.transform", text="Roll Selected", icon="TRACKING_REFINE_FORWARDS").mode='BONE_ROLL'

#-----------------------------------------------------------------------------------------------------
# Edit Armature Mode : Mirror
#-----------------------------------------------------------------------------------------------------

def _bmax_mirror_armature_execute(self, ctx):
	tmp = ctx.window.scene.transform_orientation_slots[0].type
	ctx.window.scene.transform_orientation_slots[0].type = self.c_mode
	for i, b in enumerate(ctx.selected_bones):
		b.head = self.bone_list[i][0].copy()
		b.tail = self.bone_list[i][1].copy()
		b.roll = self.bone_list[i][2]

	c_ax = (False, False, False)
	offs = (self.v_offs, self.v_offs, self.v_offs)
	if	 self.t_mode == 'X':
		c_ax = (True, False, False)
		offs = (self.v_offs, 0, 0)
	elif self.t_mode == 'Y':
		c_ax = (False, True, False)
		offs = (0, self.v_offs, 0)
	elif self.t_mode == 'Z':
		c_ax = (False, False, True)
		offs = (0, 0, self.v_offs)
	elif self.t_mode == 'XY':
		c_ax = (True, True, False)
		offs = (self.v_offs, self.v_offs, 0)
	elif self.t_mode == 'YZ':
		c_ax = (False, True, True)
		offs = (0, self.v_offs, self.v_offs)
	elif self.t_mode == 'ZX':
		c_ax = (True, False, True)
		offs = (self.v_offs, 0, self.v_offs)
	if self.t_mode != 'N':
		# mirror
		bpy.ops.transform.mirror(constraint_axis = c_ax, orient_type = self.c_mode)
		#offset
		bpy.ops.transform.translate(value = offs, orient_type = self.c_mode)

	ctx.window.scene.transform_orientation_slots[0].type = tmp

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_armature_mirror(Operator):
	bl_idname = "bmax.mirror_bone_tool"
	bl_label = "Mirror Armature"
	bl_description = "Mirror bone dialog box"
	bl_options = {'REGISTER', 'UNDO'}

	bone_list = []

	v_offs: FloatProperty(default = 0)
	t_mode: EnumProperty(
		name = 'Axis', description = 'Mirror axis', default = 'N',
		items = [
			('N', 'None', 'None'),
			('X', 'X', 'X'),
			('Y', 'Y', 'Y'),
			('Z', 'Z', 'Z'),
			('XY', 'XY', 'XY'),
			('YZ', 'YZ', 'YZ'),
			('ZX', 'ZX', 'ZX')]
		)
	c_mode: EnumProperty(
		name = 'Coord. System', description = 'Coordinate System', default = 'GLOBAL',
		items = [
			('GLOBAL', 'Global', 'Global'),
			('LOCAL', 'Local', 'Local'),
			('NORMAL', 'Normal', 'Normal')]
		)

	@classmethod
	def poll(self, ctx):
		return ctx.active_bone != None and len(ctx.selected_bones) > 0

	def check(self, ctx):
		_bmax_mirror_armature_execute(self, ctx)
		return True

	def cancel(self, ctx):
		for i, b in enumerate(ctx.selected_bones):
			b.head = self.bone_list[i][0].copy()
			b.tail = self.bone_list[i][1].copy()
			b.roll = self.bone_list[i][2]

	def draw(self, ctx):
		ui = self.layout
		b = ui.box()
		s = b.split(factor=0.35)
		c1 = s.column()
		c2 = s.column()
		c1.row().label(text="Coord. System:")
		c1.row().label(text="Axis:")
		c1.row().label(text="Offset:")
		c2.row().prop(self, "c_mode", text = "")
		c2.row().prop(self, "t_mode", text = "")
		c2.row().prop(self, "v_offs", text = "")

	def execute(self, ctx):
		return {'FINISHED'}

	def invoke(self, ctx, evt):
		self.bone_list = []
		for b in ctx.selected_bones:
			self.bone_list.append([b.head.copy(), b.tail.copy(), b.roll])
		_bmax_mirror_armature_execute(self, ctx)
		ctx.window_manager.invoke_props_dialog(self)
		return {'RUNNING_MODAL'}

#-----------------------------------------------------------------------------------------------------
# Edit Armature Mode : Clone
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_armature_clone_bones(Operator):
	bl_idname = "bmax.clone_bones"
	bl_label = "Clone To Element"
	bl_description = "Clone selection into new element"
	def execute(self, ctx):
		bpy.ops.armature.duplicate_move(ARMATURE_OT_duplicate={}, TRANSFORM_OT_translate={"value":(0, 0, 0)})
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_armature_clone(Menu):
	bl_label = "Clone/Detach"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("bmax.clone_bones", icon="MOD_BOOLEAN")
		ui.separator()
		ui.operator("armature.split", text="Detach To Element", icon="MOD_BUILD")
		ui.operator("armature.separate", text="Detach To Object", icon="MOD_BUILD")

#-----------------------------------------------------------------------------------------------------
# Edit Armature Mode : Create
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_armature_extrude(Operator):
	bl_idname = "bmax.extrude_bones"
	bl_label = "Extrude"
	bl_description = "Extrude selected bones"
	def execute(self, ctx):
		bpy.ops.armature.extrude()
		bpy.ops.wm.tool_set_by_id(name='builtin.move')
		# bpy.ops.wm.tool_set_by_id(name='builtin.extrude', cycle=False)
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------
# Edit Armature/Pose Mode : Auto-Names
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_armature_auto_names(Menu):
	bl_label = "Auto-Names"
	bl_description = "Automatic name generators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("armature.autoside_names", text="Mirror Names - Left/Right").type='XAXIS'
		ui.operator("armature.autoside_names", text="Mirror Names - Front/Back").type='YAXIS'
		ui.operator("armature.autoside_names", text="Mirror Names - Top/Bottom").type='ZAXIS'
		ui.operator("armature.flip_names", text="Flip Names")

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_pose_auto_names(Menu):
	bl_label = "Auto-Names"
	bl_description = "Automatic name generators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("pose.autoside_names", text="Mirror Names - Left/Right").axis='XAXIS'
		ui.operator("pose.autoside_names", text="Mirror Names - Front/Back").axis='YAXIS'
		ui.operator("pose.autoside_names", text="Mirror Names - Top/Bottom").axis='ZAXIS'
		ui.operator("pose.flip_names", text="Flip Names")

#-----------------------------------------------------------------------------------------------------
# Edit Armature/Pose Mode : Hide/Un-hide
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_armature_hide(Menu):
	bl_label = "Hide/Unhide"
	bl_description = "Hide/Unhide operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("armature.hide", text="Hide Selected").unselected=False
		ui.operator("armature.hide", text="Hide Unselected").unselected=True
		ui.separator()
		ui.operator("armature.reveal", text="Unhide All")

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_pose_hide(Menu):
	bl_label = "Hide/Unhide"
	bl_description = "Hide/Unhide operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("pose.hide", text="Hide Selected").unselected=False
		ui.operator("pose.hide", text="Hide Unselected").unselected=True
		ui.separator()
		ui.operator("pose.reveal", text="Unhide All")

#-----------------------------------------------------------------------------------------------------
# Edit Pose Mode : Pose Library
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_pose_tools(Menu):
	bl_label = "Poses"
	bl_description = "Pose tools"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("pose.copy", text="Copy", icon="COPYDOWN")
		ui.operator("pose.paste", text="Paste", icon="PASTEDOWN").flipped=False
		ui.operator("pose.paste", text="Paste Opposite", icon="PASTEFLIPDOWN").flipped=True
		ui.separator()
		ui.operator("poselib.pose_add", text="Save To Library", icon="ADD")
		ui.operator("poselib.pose_remove", text="Delete From Library", icon="REMOVE")
		ui.operator("poselib.pose_rename", text="Rename")
		ui.operator("poselib.browse_interactive", text="Browse Library")

#-----------------------------------------------------------------------------------------------------
# Edit Object/Pose Mode : Constraints
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_object_constraints(Menu):
	bl_label = "Animation Constraints"
	bl_description = "Animation constraint tools"
	def draw(self, ctx):
		ui = self.layout
		ui.operator_menu_enum("object.constraint_add", "type", text="Add (With Targets)", icon="ADD")
		ui.operator("object.constraints_clear", text="Clear Constraints", icon="REMOVE")
		ui.operator("object.constraints_copy", text="Copy To Selected", icon="ARROW_LEFTRIGHT")

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_pose_constraints(Menu):
	bl_label = "Animation Constraints"
	bl_description = "Animation constraint tools"
	def draw(self, ctx):
		ui = self.layout
		ui.operator_menu_enum("pose.constraint_add", "type", text="Add (With Targets)", icon="ADD")
		ui.operator("pose.constraints_clear", text="Clear Constraints", icon="REMOVE")
		ui.operator("pose.constraints_copy", text="Copy To Selected", icon="ARROW_LEFTRIGHT")

#-----------------------------------------------------------------------------------------------------
# Curve Edit Mode : Show/Hide Selection
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_curve_hide(Menu):
	bl_label = "Hide/Unhide"
	bl_description = "Hide/unhide operators"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("curve.hide", text = "Hide Selected").unselected = False
		ui.operator("curve.hide", text = "Hide Unselected").unselected = True
		ui.separator()
		ui.operator("curve.reveal", text = "Unhide All")

#-----------------------------------------------------------------------------------------------------
# Curve Edit Mode : Clone
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_curve_clone(Operator):
	bl_idname = "bmax.clone_curve"
	bl_label = "Clone To Element"
	bl_description = "Clone selection into new element"
	def execute(self, ctx):
		bpy.ops.curve.duplicate_move(CURVE_OT_duplicate={}, TRANSFORM_OT_translate={"value":(0, 0, 0)})
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------
# Object/Pose Mode : Keyframing Tools
#-----------------------------------------------------------------------------------------------------

class BMAX_PT_keyframing_settings(Panel):
	bl_label = "Keyframing Settings"
	bl_options = {'HIDE_HEADER'}
	bl_region_type = 'HEADER'
	bl_space_type = 'VIEW_3D'

	def draw(self, ctx):
		ui = self.layout

		ui.row().label(text="Active Keying Set:")
		ui.row().prop_search(ctx.scene.keying_sets_all, "active", ctx.scene, "keying_sets_all", text="")

		ui.row().label(text="New Keyframe Type:")
		ui.row().prop(ctx.tool_settings, "keyframe_type", text="")

		c = ui.column(align=True)
		c.label(text="Auto Keyframing:")
		r = c.row()
		r.prop(ctx.tool_settings, "auto_keying_mode", text="")
		r.prop(ctx.tool_settings, "use_keyframe_insert_keyingset", text="")
		
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_object_keyframing(Menu):
	bl_label = "Animation Tools"
	bl_description = "Animation tools"
	def draw(self, ctx):
		ui = self.layout
		if ctx.mode == 'OBJECT':
			ui.menu("BMAX_MT_object_constraints", icon="CONSTRAINT")
		else:
			ui.menu("BMAX_MT_pose_constraints", icon="CONSTRAINT")
		ui.popover(panel="BMAX_PT_keyframing_settings", text="Keyframing Settings", icon="SETTINGS")
		ui.separator()
		ui.operator("anim.keyframe_clear_v3d", text="Delete Animation")

#-----------------------------------------------------------------------------------------------------
# All Modes : Viewport Settings
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_viewport(Menu):
	bl_label = "3D Viewport Settings"
	bl_description = "3D Viewport settings"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("view3d.view_persportho", text="Perspective Toggle", icon = "VIEW_PERSPECTIVE")
		ui.separator()
		ui.operator("view3d.view_axis", text = "Front Ortho", icon = "AXIS_FRONT").type='FRONT'
		ui.operator("view3d.view_axis", text = "Left Ortho", icon = "AXIS_SIDE").type='LEFT'
		ui.operator("view3d.view_axis", text = "Top Ortho", icon = "AXIS_TOP").type='TOP'
		ui.separator()
		ui.operator("view3d.view_axis", text = "Back Ortho", icon = "AXIS_FRONT").type='BACK'
		ui.operator("view3d.view_axis", text = "Right Ortho", icon = "AXIS_SIDE").type='RIGHT'
		ui.operator("view3d.view_axis", text = "Bottom Ortho", icon = "AXIS_TOP").type='BOTTOM'
		ui.separator()
		if ctx.screen.show_fullscreen:
			ui.operator("screen.screen_full_area", text="Restore Viewport", icon="FULLSCREEN_EXIT")
		else:
			ui.operator("screen.screen_full_area", text="Maximize Viewport", icon="FULLSCREEN_ENTER")

#-----------------------------------------------------------------------------------------------------
# Blender 3ds Max tools emulation : Right-Click Menu
#-----------------------------------------------------------------------------------------------------

def _bmax_mesh_subobject(mode):
	t = "Edit Custom"
	i = "SNAP_FACE"
	if mode[0] and mode [1] and not mode[2]:
		t = "Edit Edges/Vertices"
		i = "SNAP_VERTEX"
	elif mode[0] and not mode[1] and not mode[2]:
		t = "Edit Vertices"
		i = "VERTEXSEL"
	elif mode[1] and not mode[0] and not mode[2]:
		t = "Edit Edges"
		i = "EDGESEL"
	elif mode[2] and not mode[1] and not mode[0]:
		t = "Edit Faces"
		i = "FACESEL"
	return t, i

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_tool_menu(Menu):
	bl_label = "BMax Tools"
	bl_description = "BMax right-click menu"

	def draw(self, ctx):
		ui = self.layout
		ui.operator_context = 'INVOKE_REGION_WIN'

		ui.operator("screen.redo_last", text = "Edit Last Action...", icon = "MENU_PANEL")
		ui.separator()

		#create
		if ctx.mode == 'OBJECT':
			ui.menu("VIEW3D_MT_add", text="Create", icon="SOLO_OFF")
		elif ctx.mode == 'EDIT_MESH':
			ui.menu("VIEW3D_MT_mesh_add", text="Create", icon='SOLO_OFF')
		elif ctx.mode == 'EDIT_CURVE':
			ui.menu("VIEW3D_MT_curve_add", text="Create", icon='SOLO_OFF')
		elif ctx.mode == 'EDIT_ARMATURE':
			ui.menu("BMAX_MT_armature_create_primitive", icon='SOLO_OFF')

		if ctx.mode == 'OBJECT':
			ui.menu("BMAX_MT_object_hide", icon="RESTRICT_VIEW_OFF")
			ui.menu("BMAX_MT_object_freeze", icon="FREEZE")
			ui.separator()
			if ctx.active_object and ctx.active_object.type in {'MESH', 'CURVE', 'SURFACE', 'META', 'FONT', 'ARMATURE', 'LATTICE'}:
				ui.operator("object.mode_set", text="Edit Object", icon="EDITMODE_HLT").mode='EDIT'
				if ctx.active_object.type == 'ARMATURE':
					ui.operator("object.mode_set", text="Edit Pose", icon="POSE_HLT").mode='POSE'
			else:
				ui.operator("object.mode_set", text="Edit Object", icon="EDITMODE_HLT")
			ui.separator()
			ui.menu("BMAX_MT_object_link", icon="LINK_BLEND")
			ui.operator("object.join", text="Attach")
			ui.separator()
			ui.operator("bmax.align_tool", text="Align Selection...", icon="CENTER_ONLY")
			ui.menu("BMAX_MT_object_align_special", icon="CENTER_ONLY")
			ui.operator("bmax.mirror_tool", text="Mirror...", icon="MOD_MIRROR")
			ui.operator("bmax.clone_tool", text="Clone Options...", icon="MOD_BOOLEAN")
			ui.separator()
			op = ui.operator("object.transform_apply", text="Reset XForm")
			op.rotation = True
			op.scale = True
			ui.operator("bmax.check_valid_faces", text="Select N-Gons")
			ui.separator()

		# edit mesh
		elif ctx.mode == 'EDIT_MESH':
			t, i = _bmax_mesh_subobject(ctx.tool_settings.mesh_select_mode)
			ui.menu("BMAX_MT_mesh_sub_object", icon=i)
			ui.menu("BMAX_MT_mesh_show_hide_element", icon="RESTRICT_VIEW_OFF")
			ui.separator()
			ui.menu("BMAX_MT_mesh_select_special", icon="RESTRICT_SELECT_OFF")
			ui.separator()
			ui.menu("BMAX_MT_mesh_align_special", icon="CENTER_ONLY")
			ui.menu("BMAX_MT_mesh_clone", icon="MOD_BOOLEAN")
			ui.separator()
			ui.menu("BMAX_MT_mesh_create", icon="SNAP_FACE")
			ui.menu("BMAX_MT_mesh_slice", icon="MOD_DECIM")
			ui.menu("VIEW3D_MT_edit_mesh_extrude", icon="CURVE_PATH")
			ui.menu("BMAX_MT_mesh_repair", icon="MODIFIER")
			ui.separator()
			#tools
			ui.menu("BMAX_MT_mesh_meshtools", icon="GRID")
			if _bmax_weld_check(ctx):
				ui.operator("bmax.target_weld_off", icon="SNAP_ON")
			else:
				if ctx.tool_settings.use_mesh_automerge == True:
					ui.operator("bmax.target_weld_off", icon="AUTOMERGE_ON")
				else:
					ui.operator("bmax.target_weld_on", icon="SNAP_OFF")
			ui.separator()
			###
			ui.menu("VIEW3D_MT_uv_map", icon="MOD_UVPROJECT")
			ui.operator("mesh.mark_seam", text="Set Seams")
			ui.operator("mesh.mark_seam", text="Clear Seams").clear=True
			ui.separator()

		# edit armature
		elif ctx.mode == 'EDIT_ARMATURE':
			ui.menu("BMAX_MT_armature_hide", icon="RESTRICT_VIEW_OFF")
			ui.menu("BMAX_MT_armature_auto_names", icon="PREFERENCES")
			ui.separator()
			ui.operator("object.mode_set", text="Edit Pose", icon="POSE_HLT").mode='POSE'
			ui.separator()
			ui.operator("armature.parent_set", text="Link", icon = "LINKED")
			ui.operator("armature.parent_clear", text="Unlink", icon = "UNLINKED")
			ui.separator()
			ui.menu("BMAX_MT_armature_align_special", icon="CENTER_ONLY")
			ui.menu("BMAX_MT_armature_clone", icon="MOD_BOOLEAN")
			ui.operator("bmax.mirror_bone_tool", text="Mirror...", icon="MOD_MIRROR") 
			ui.separator()
			ui.menu("BMAX_MT_armature_bone_tools", icon="GROUP_BONE")
			ui.operator("bmax.extrude_bones", text="Extrude", icon="CURVE_PATH")
			ui.operator("armature.switch_direction", text="Flip Direction")
			ui.separator()

		# edit pose
		elif ctx.mode == 'POSE':
			ui.menu("BMAX_MT_pose_hide", icon="RESTRICT_VIEW_OFF")
			ui.menu("BMAX_MT_pose_auto_names", icon="PREFERENCES")
			ui.separator()
			ui.operator("object.mode_set", text="Edit Bones", icon="EDITMODE_HLT").mode='EDIT'
			ui.separator()
			ui.operator("object.parent_set", text="Link", icon="LINKED")
			ui.operator("object.parent_clear", text="Unlink", icon="UNLINKED")
			ui.separator()
			ui.menu("BMAX_MT_armature_align_special", icon="CENTER_ONLY")
			ui.separator()
			ui.operator("pose.armature_apply", text="Freeze Transform", icon="FREEZE")
			ui.operator("pose.transforms_clear", text="Transform To Zero", icon="OUTLINER_DATA_ARMATURE")
			ui.operator("pose.rot_clear", text="Rotation To Zero", icon="DRIVER_ROTATIONAL_DIFFERENCE")
			ui.operator("pose.quaternions_flip", text="Flip Rotation")
			ui.separator()
			ui.menu("BMAX_MT_pose_tools", icon="ARMATURE_DATA")
			ui.separator()

		# edit curve
		elif ctx.mode == 'EDIT_CURVE':
			ui.menu("BMAX_MT_curve_hide", icon="RESTRICT_VIEW_OFF")
			ui.separator()
			ui.menu("VIEW3D_MT_snap", text="Align Special", icon="CENTER_ONLY")
			ui.operator("transform.mirror", text="Mirror", icon="MOD_MIRROR")
			ui.operator("bmax.clone_curve", text="Clone To Element", icon="MOD_BOOLEAN")
			ui.operator("curve.separate", text="Detach To Object", icon="MOD_BUILD")
			ui.separator()
			ui.operator("curve.make_segment", text="Create Segment", icon="SPHERECURVE")
			ui.operator("curve.handle_type_set", icon="CURVE_BEZCURVE")
			ui.operator("curve.subdivide")
			ui.operator("curve.switch_direction", text="Reverse Direction")
			ui.operator("transform.tilt", icon="TRACKING_REFINE_FORWARDS")
			ui.operator("curve.tilt_clear", icon="TRACKING_REFINE_FORWARDS")
			ui.separator()

		if ctx.mode in {'OBJECT', 'POSE'}:
			ui.operator("anim.keyframe_insert", text="Insert Keyframe", icon="KEY_HLT")
			ui.operator("anim.keyframe_delete_v3d", text="Delete Keyframe", icon="KEY_DEHLT")
			ui.menu("BMAX_MT_object_keyframing", icon="ANIM")
			ui.separator()

		if ctx.mode != 'OBJECT':
			ui.operator("object.mode_set", text="Exit Edit Mode", icon="OBJECT_DATAMODE").mode='OBJECT'
			ui.separator()

		if ctx.mode in {'OBJECT', 'EDIT_MESH', 'EDIT_ARMATURE', 'POSE', 'EDIT_CURVE'}:
			ui.operator("transform.select_orientation", text="Coordinate System", icon="OBJECT_ORIGIN")
			ui.operator("wm.context_menu_enum", text="Rotation/Scale Pivot", icon="PIVOT_BOUNDBOX").data_path='tool_settings.transform_pivot_point'

		ui.menu("BMAX_MT_viewport", icon="FCURVE")

#-----------------------------------------------------------------------------------------------------
#
#							          UV Right-Click Menu
#
#-----------------------------------------------------------------------------------------------------

# edit vertices
class BMAX_OT_uv_sub_object_verts(Operator):
	bl_idname = "bmax.sub_object_verts_uv"
	bl_label = "Edit Vertices"
	bl_description = "Select vertex sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.uv_select_mode = 'VERTEX'
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

# edit edges
class BMAX_OT_uv_sub_object_edges(Operator):
	bl_idname = "bmax.sub_object_edges_uv"
	bl_label = "Edit Edges"
	bl_description = "Select edge sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.uv_select_mode = 'EDGE'
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

# edit faces
class BMAX_OT_uv_sub_object_faces(Operator):
	bl_idname = "bmax.sub_object_faces_uv"
	bl_label = "Edit Faces"
	bl_description = "Select face sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.uv_select_mode = 'FACE'
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

# edit islands
class BMAX_OT_uv_sub_object_islands(Operator):
	bl_idname = "bmax.sub_object_islands_uv"
	bl_label = "Edit Islands"
	bl_description = "Select island sub-object level"
	def execute(self, ctx):
		ctx.tool_settings.uv_select_mode = 'ISLAND'
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

# Sub-Menu for last 4 operators
class BMAX_MT_uv_sub_object(Menu):
	bl_label = "Sub-Object"
	bl_description = "Select sub-object level"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("bmax.sub_object_verts_uv", icon = 'UV_VERTEXSEL')
		ui.operator("bmax.sub_object_edges_uv", icon = 'UV_EDGESEL')
		ui.operator("bmax.sub_object_faces_uv", icon = 'UV_FACESEL')
		ui.operator("bmax.sub_object_islands_uv", icon = 'UV_ISLANDSEL')

#-----------------------------------------------------------------------------------------------------
# Show/Hide Selection
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_uv_show_hide_element(Menu):
	bl_label = "Hide/Unhide"
	bl_description = "Hide/Unhide sub-object selection"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("uv.hide", text = "Hide Selected").unselected = False
		ui.operator("uv.hide", text = "Hide Unselected").unselected = True
		ui.separator()
		ui.operator("uv.reveal", text = "Unhide All")

#-----------------------------------------------------------------------------------------------------
# Space UV-s
# based on code by (c) 2012 Dan Wheeler
#-----------------------------------------------------------------------------------------------------

def _bmax_distribute_uv(ctx, dox, doy):
	#collect selection
	sel = {}
	obj = ctx.active_object
	uvw = obj.data.uv_layers.active
	for loop in obj.data.loops:
		if uvw.data[loop.index].select:
			u = uvw.data[loop.index].uv[0]
			v = uvw.data[loop.index].uv[1]
			uv = u, v
			if not uv in sel:
				sel[uv] = []
			sel[uv].append(loop.index)
	#space UVs
	if len(sel) >= 3:
		sx = []
		sy = []
		selx = {}
		sely = {}
		for k in sel.keys():
			sx.append(k[0])
			sy.append(k[1])
		sx.sort()
		sy.sort()
		dx = sx[-1] - sx[0]
		dy = sy[-1] - sy[0]
		i=0
		for x in sx:
			selx[x] = i
			i += 1
		i=0
		for y in sy:
			sely[y] = i
			i += 1
		szx = len(selx) - 1
		szy = len(sely) - 1
		for k in sel.keys():
			u = k[0]
			v = k[1]
			x = u
			y = v
			if szx > 0 and dox == True: x = (selx[u] / szx) * dx + sx[0]
			if szy > 0 and doy == True: y = (sely[v] / szy) * dy + sy[0]
			a = sel[k]
			for id in a:
				uvw.data[id].uv[0] = x
				uvw.data[id].uv[1] = y

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_uv_space_horizontal(Operator):
	bl_idname = "bmax.space_horizontal_uv"
	bl_label = "Space Horizontal"
	bl_description = "Evenly space UVs horizontally"
	bl_options = {'UNDO'}
	def execute(self, ctx):
		bpy.ops.object.editmode_toggle()
		_bmax_distribute_uv(ctx, True, False)
		bpy.ops.object.editmode_toggle()
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_uv_space_vertical(Operator):
	bl_idname = "bmax.space_vertical_uv"
	bl_label = "Space Vertical"
	bl_description = "Evenly space UVs vertically"
	bl_options = {'UNDO'}
	def execute(self, ctx):
		bpy.ops.object.editmode_toggle()
		_bmax_distribute_uv(ctx, False, True)
		bpy.ops.object.editmode_toggle()
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------
# Rotate Selection
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_uv_rotate_plus(Operator):
	bl_idname = "bmax.rotate_plus_uv"
	bl_label = "Rotate 90 CW"
	bl_description = "Rotate the selection 90 degrees"
	def execute(self, ctx):
		bpy.ops.transform.rotate(
			value=1.5708, 
			orient_axis='Z', 
			orient_type='VIEW', 
			orient_matrix=((-1, -0, -0), (-0, -1, -0), (-0, -0, -1)), 
			orient_matrix_type='VIEW', 
			mirror=True)
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_uv_rotate_minus(Operator):
	bl_idname = "bmax.rotate_minus_uv"
	bl_label = "Rotate 90 CCW"
	bl_description = "Rotate the selection -90 degrees"
	def execute(self, ctx):
		bpy.ops.transform.rotate(
			value=-1.5708,
			orient_axis='Z', 
			orient_type='VIEW', 
			orient_matrix=((-1, -0, -0), (-0, -1, -0), (-0, -0, -1)), 
			orient_matrix_type='VIEW', 
			mirror=True)
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------
# Flip Selection
#-----------------------------------------------------------------------------------------------------

class BMAX_OT_uv_flip_horizontal(Operator):
	bl_idname = "bmax.flip_horizontal_uv"
	bl_label = "Flip Horizontal"
	bl_description = "Detach edge verts and mirror horizontally"
	def execute(self, ctx):
		if ctx.tool_settings.use_uv_select_sync == False:
			bpy.ops.uv.select_split()
		bpy.ops.transform.mirror(constraint_axis=(True, False, False), orient_type='GLOBAL')
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_uv_flip_vertical_uv(Operator):
	bl_idname = "bmax.flip_vertical_uv"
	bl_label = "Flip Vertical"
	bl_description = "Detach edge verts and mirror vertically"
	def execute(self, ctx):
		if ctx.tool_settings.use_uv_select_sync == False:
			bpy.ops.uv.select_split()
		bpy.ops.transform.mirror(constraint_axis=(False, True, False), orient_type='GLOBAL')
		return {'FINISHED'}
	def invoke(self, ctx, evt):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------
# UV Tools
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_uv_tools(Menu):
	bl_label = "Tools"
	bl_description = "UV Tools"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("uv.weld", text = "Weld Selected", icon = "MOD_DISPLACE")
		ui.operator("uv.stitch", text = "Stitch Selected")
		ui.separator()
		ui.operator("uv.select_split", text = "Detach Edge Verts")
		ui.separator()
		ui.operator("uv.unwrap", icon = "MOD_UVPROJECT")
		ui.operator("uv.mark_seam", text = "Set Seams").clear=False
		ui.operator("uv.mark_seam", text = "Clear Seams").clear=True
		ui.separator()
		ui.operator("uv.pack_islands", text = "Pack UVs")
		ui.operator("uv.minimize_stretch", text = "Relax")
		ui.separator()
		ui.operator("uv.export_layout", text = "Render UVW Template", icon = "TEXTURE")

#-----------------------------------------------------------------------------------------------------
# BMax UV Tools Right-Click Menu
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_tool_menu_uv(Menu):
	bl_label = "BMax UV Tools"
	bl_description = "BMax right-click menu"

	def draw(self, ctx):
		ui = self.layout
		ui.operator_context = 'INVOKE_REGION_WIN'

		ui.operator("screen.redo_last", text = "Edit Last Action...", icon = "MENU_PANEL")
		ui.separator()
		if ctx.tool_settings.use_uv_select_sync == True:
			ui.menu("BMAX_MT_mesh_sub_object")
		else:
			ui.menu("BMAX_MT_uv_sub_object")
		ui.menu("BMAX_MT_uv_show_hide_element", icon="RESTRICT_VIEW_OFF")
		ui.prop(ctx.tool_settings, "use_uv_select_sync", text="Sync Selection")
		ui.separator()
		ui.operator("uv.select_linked", text="Select Element", icon="UV_ISLANDSEL")
		ui.separator()
		ui.menu("IMAGE_MT_uvs_snap", text = "Align Selection", icon = "CENTER_ONLY")
		ui.operator("uv.align", text = "Align Vertical").axis='ALIGN_X'
		ui.operator("uv.align", text = "Align Horizontal").axis='ALIGN_Y'
		ui.separator()
		ui.operator("bmax.space_vertical_uv")
		ui.operator("bmax.space_horizontal_uv")
		ui.separator()
		ui.operator("bmax.rotate_plus_uv")
		ui.operator("bmax.rotate_minus_uv")
		ui.operator("bmax.flip_vertical_uv")
		ui.operator("bmax.flip_horizontal_uv")
		ui.separator()
		ui.menu("BMAX_MT_uv_tools", icon = "MOD_UVPROJECT")
		ui.separator()
		ui.operator("wm.context_menu_enum", text="Rotation/Scale Pivot", icon="PIVOT_BOUNDBOX").data_path='space_data.pivot_point'
		if ctx.screen.show_fullscreen:
			ui.operator("screen.screen_full_area", text="Restore Viewport", icon="FULLSCREEN_EXIT")
		else:
			ui.operator("screen.screen_full_area", text="Maximize Viewport", icon="FULLSCREEN_ENTER")

#-----------------------------------------------------------------------------------------------------
#
#							     Dopesheet & Graph Editor
#
#-----------------------------------------------------------------------------------------------------

class BMAX_MT_graph_align_special(Menu):
	bl_label = "Align Special"
	bl_description = "Align selection to specified point"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("graph.snap", text = "Flatten Handles").type='HORIZONTAL'
		ui.separator()
		ui.operator("graph.snap", text = "Selection To Cursor Value").type='VALUE'
		ui.operator("graph.snap", text = "Selection To Cursor Frame").type='CFRA'
		ui.operator("graph.snap", text = "Selection To Nearest Frame").type='NEAREST_FRAME'
		ui.operator("graph.snap", text = "Selection To Nearest Second").type='NEAREST_SECOND'
		ui.operator("graph.snap", text = "Selection To Nearest Marker").type='NEAREST_MARKER'
		ui.separator()
		ui.operator("graph.frame_jump", text = "Cursor To Selection")

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_tool_menu_graph(Menu):
	bl_label = "BMax Graph Tools"
	bl_description = "BMax right-click menu"

	def draw(self, ctx):
		ui = self.layout
		ui.operator_context = 'INVOKE_REGION_WIN'

		ui.operator("screen.redo_last", text = "Edit Last Action...", icon = "MENU_PANEL")
		ui.separator()
		ui.operator("graph.copy", text="Copy", icon='COPYDOWN')
		ui.operator("graph.paste", text="Paste", icon='PASTEDOWN')
		ui.operator("graph.paste", text="Paste Flipped", icon='PASTEFLIPDOWN').flipped = True
		ui.separator()
		# manipulation
		ui.menu("BMAX_MT_graph_align_special", icon="CENTER_ONLY")
		ui.operator_menu_enum("graph.mirror", "type", text="Mirror", icon="MOD_MIRROR")
		ui.operator("graph.duplicate_move", text = "Clone And Move", icon="MOD_BOOLEAN")
		ui.separator()
		# navigation
		ui.operator("screen.keyframe_jump", text="Next Frame", icon="FRAME_NEXT").next=True
		ui.operator("screen.keyframe_jump", text="Previous Frame", icon="FRAME_PREV").next=False
		# ui.operator("screen.frame_jump", text="Go To Start", icon="TRIA_LEFT_BAR").end=False
		ui.separator()
		# keyframing
		if ctx.mode in {'OBJECT', 'POSE'}:
			ui.operator("anim.keyframe_insert", text="Insert Keyframe", icon="KEY_HLT")
			ui.operator("anim.keyframe_delete_v3d", text="Delete Keyframe", icon="KEY_DEHLT")
		ui.operator("anim.keyframe_clear_v3d", text="Delete Animation")
		ui.separator()
		# settings
		ui.operator_menu_enum("graph.handle_type", "type", text = "Handle Type", icon="CURVE_BEZCURVE")
		ui.operator_menu_enum("graph.easing_type", "type", text = "Easing Type", icon="IPO_EASE_IN_OUT")
		ui.operator_menu_enum("graph.interpolation_type", "type", text = "Interpolation", icon="IPO_CONSTANT")
		ui.separator()
		# tools
		op = ui.operator("wm.context_set_enum", text = "Switch To Dopesheet", icon="ACTION")
		op.data_path = 'area.type'
		op.value = 'DOPESHEET_EDITOR'
		ui.separator()
		ui.operator("wm.context_menu_enum", text="Rotation/Scale Pivot", icon="PIVOT_BOUNDBOX").data_path='space_data.pivot_point'
		if ctx.screen.show_fullscreen:
			ui.operator("screen.screen_full_area", text="Restore Viewport", icon="FULLSCREEN_EXIT")
		else:
			ui.operator("screen.screen_full_area", text="Maximize Viewport", icon="FULLSCREEN_ENTER")

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_dopesheet_align_special(Menu):
	bl_label = "Align Special"
	bl_description = "Align selection to specified point"
	def draw(self, ctx):
		ui = self.layout
		ui.operator("action.snap", text = "Selection To Cursor Frame").type='CFRA'
		ui.operator("action.snap", text = "Selection To Nearest Frame").type='NEAREST_FRAME'
		ui.operator("action.snap", text = "Selection To Nearest Second").type='NEAREST_SECOND'
		ui.operator("action.snap", text = "Selection To Nearest Marker").type='NEAREST_MARKER'
		ui.separator()
		ui.operator("action.frame_jump", text = "Cursor To Selection")

#-----------------------------------------------------------------------------------------------------

class BMAX_MT_tool_menu_dopesheet(Menu):
	bl_label = "BMax Dopesheet Tools"
	bl_description = "BMax right-click menu"

	def draw(self, ctx):
		ui = self.layout
		ui.operator_context = 'INVOKE_REGION_WIN'

		ui.operator("screen.redo_last", text = "Edit Last Action...", icon = "MENU_PANEL")
		ui.separator()
		ui.operator("action.copy", text="Copy", icon='COPYDOWN')
		ui.operator("action.paste", text="Paste", icon='PASTEDOWN')
		ui.operator("action.paste", text="Paste Flipped", icon='PASTEFLIPDOWN').flipped = True
		ui.separator()
		# manipulation
		ui.menu("BMAX_MT_dopesheet_align_special", icon="CENTER_ONLY")
		ui.operator_menu_enum("action.mirror", "type", text="Mirror", icon="MOD_MIRROR")
		ui.operator("action.duplicate_move", text = "Clone And Move", icon="MOD_BOOLEAN")
		ui.separator()
		# navigation
		ui.operator("screen.keyframe_jump", text="Next Frame", icon="FRAME_NEXT").next=True
		ui.operator("screen.keyframe_jump", text="Previous Frame", icon="FRAME_PREV").next=False
		# ui.operator("screen.frame_jump", text="Go To Start", icon="TRIA_LEFT_BAR").end=False
		ui.separator()
		# keyframing
		if ctx.mode in {'OBJECT', 'POSE'}:
			ui.operator("anim.keyframe_insert", text="Insert Keyframe", icon="KEY_HLT")
			ui.operator("anim.keyframe_delete_v3d", text="Delete Keyframe", icon="KEY_DEHLT")
		ui.operator("anim.keyframe_clear_v3d", text="Delete Animation")
		ui.separator()
		# settings
		# ui.operator_menu_enum("action.handle_type", "type", text = "Handle Type", icon="CURVE_BEZCURVE")
		ui.operator_menu_enum("action.keyframe_type", "type", text = "Keyframe Type", icon="KEYFRAME")
		ui.operator_menu_enum("action.interpolation_type", "type", text = "Interpolation", icon="IPO_CONSTANT")
		# ui.operator_menu_enum("action.easing_type", "type", text="Easing Mode", icon="IPO_EASE_IN_OUT")
		ui.separator()
		# tools
		op = ui.operator("wm.context_set_enum", text = "Switch To Graph", icon="GRAPH")
		op.data_path = 'area.type'
		op.value = 'GRAPH_EDITOR'
		ui.separator()
		if ctx.screen.show_fullscreen:
			ui.operator("screen.screen_full_area", text="Restore Viewport", icon="FULLSCREEN_EXIT")
		else:
			ui.operator("screen.screen_full_area", text="Maximize Viewport", icon="FULLSCREEN_ENTER")


#-----------------------------------------------------------------------------------------------------
#
#						         Shading previews
#
#-----------------------------------------------------------------------------------------------------

def _bmax_get_shading(ctx):
	view = ctx.space_data
	if view.type == 'VIEW_3D':
		return view.shading
	else:
		return ctx.scene.display.shading

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_preview_default(Operator):
	bl_idname = "bmax.preview_default"
	bl_label = "Default"
	bl_description = "Setup default preview"

	@classmethod
	def poll(cls, ctx):
		return _bmax_get_shading(ctx).type in {'SOLID', 'WIREFRAME'}

	def execute(self, ctx):
		shading = _bmax_get_shading(ctx)
		if shading.type == 'SOLID':
			shading.light = 'STUDIO'
			shading.color_type = 'TEXTURE'
			shading.background_type = 'THEME'
		else:
			shading.wireframe_color_type = 'SINGLE'
			shading.background_type = 'THEME'
		return {'FINISHED'}

	def invoke(self, ctx, _event):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_preview_black_on_white(Operator):
	bl_idname = "bmax.preview_black_on_white"
	bl_label = "Silhouette B"
	bl_description = "Setup black on white silhouette preview"

	@classmethod
	def poll(cls, ctx):
		return _bmax_get_shading(ctx).type == 'SOLID'

	def execute(self, ctx):
		shading = _bmax_get_shading(ctx)
		shading.light = 'FLAT'
		shading.color_type = 'SINGLE'
		shading.single_color = (0.0, 0.0, 0.0)
		shading.background_type = 'VIEWPORT'
		shading.background_color = (0.924, 0.924, 0.924)
		return {'FINISHED'}

	def invoke(self, ctx, _event):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_preview_white_on_black(Operator):
	bl_idname = "bmax.preview_white_on_black"
	bl_label = "Silhouette W"
	bl_description = "Setup white on black silhouette preview"

	@classmethod
	def poll(cls, ctx):
		return _bmax_get_shading(ctx).type == 'SOLID'

	def execute(self, ctx):
		shading = _bmax_get_shading(ctx)
		shading.light = 'FLAT'
		shading.color_type = 'SINGLE'
		shading.single_color = (0.924, 0.924, 0.924)
		shading.background_type = 'VIEWPORT'
		shading.background_color = (0.0, 0.0, 0.0)
		return {'FINISHED'}

	def invoke(self, ctx, _event):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_preview_elsyiun(Operator):
	bl_idname = "bmax.preview_elsyiun"
	bl_label = "Elsyiun"
	bl_description = "Set Elsyiun background"

	@classmethod
	def poll(cls, ctx):
		return _bmax_get_shading(ctx).type in {'SOLID', 'WIREFRAME'}

	def execute(self, ctx):
		shading = _bmax_get_shading(ctx)
		shading.background_type = 'VIEWPORT'
		shading.background_color = (0.07036, 0.07036, 0.07036)
		return {'FINISHED'}

	def invoke(self, ctx, _event):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_preview_softimage(Operator):
	bl_idname = "bmax.preview_softimage"
	bl_label = "XSI"
	bl_description = "Set XSI background"

	@classmethod
	def poll(cls, ctx):
		return _bmax_get_shading(ctx).type in {'SOLID', 'WIREFRAME'}

	def execute(self, ctx):
		shading = _bmax_get_shading(ctx)
		shading.background_type = 'VIEWPORT'
		shading.background_color = (0.21223, 0.21223, 0.21223)
		return {'FINISHED'}

	def invoke(self, ctx, _event):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

class BMAX_OT_preview_maya_light(Operator):
	bl_idname = "bmax.preview_maya_light"
	bl_label = "Maya Light"
	bl_description = "Set Maya light background"

	@classmethod
	def poll(cls, ctx):
		return _bmax_get_shading(ctx).type in {'SOLID', 'WIREFRAME'}

	def execute(self, ctx):
		shading = _bmax_get_shading(ctx)
		shading.background_type = 'VIEWPORT'
		shading.background_color = (0.36131, 0.36131, 0.36131)
		return {'FINISHED'}

	def invoke(self, ctx, _event):
		return self.execute(ctx)

#-----------------------------------------------------------------------------------------------------

def bmax_shading_extension(self, ctx):
	ui = self.layout
	ui.separator()
	c = ui.column(align=True)
	c.row().label(text="Preview Setup")
	r = c.row(align=True)
	r.operator("bmax.preview_default")
	r.operator("bmax.preview_black_on_white")
	r.operator("bmax.preview_white_on_black")
	r = c.row(align=True)
	r.operator("bmax.preview_elsyiun")
	r.operator("bmax.preview_softimage")
	r.operator("bmax.preview_maya_light")

#-----------------------------------------------------------------------------------------------------
#
#						         3D view tool tracker
#                 (preserves tool selection across modes and workspaces)
#
#-----------------------------------------------------------------------------------------------------

from bpy.app.handlers import persistent

bmax_tool_store = None

bmax_mode_owner = object()
bmax_tool_owner = object()
bmax_work_owner = object()

#-----------------------------------------------------------------------------------------------------

def _bmax_on_tool_change(*args):
	global bmax_tool_store
	bmax_tool_store = bpy.context.workspace.tools.from_space_view3d_mode(bpy.context.mode).idname

def _bmax_on_mode_change(*args):
	if bmax_tool_store is not None:
		bpy.ops.wm.tool_set_by_id(name=bmax_tool_store, space_type='VIEW_3D')

def _bmax_on_work_change(*args):
	if bmax_tool_store is not None:
		bpy.ops.wm.tool_set_by_id(name=bmax_tool_store, space_type='VIEW_3D')

def _bmax_on_look_change(*args):
	if bmax_tool_store is not None:
		bpy.ops.wm.tool_set_by_id(name=bmax_tool_store, space_type='VIEW_3D')

#-----------------------------------------------------------------------------------------------------

def _bmax_subscribe_tool_tracker():
	rna_list = [
		((bpy.types.Object, "mode"), bmax_mode_owner, _bmax_on_mode_change),
		((bpy.types.WorkSpace, "tools"), bmax_tool_owner, _bmax_on_tool_change),
		((bpy.types.Window, "workspace"), bmax_work_owner, _bmax_on_work_change),
	]
	for rna in rna_list:
		bpy.msgbus.subscribe_rna(key=rna[0], owner=rna[1], args=(), notify=rna[2], options={'PERSISTENT'})

#-----------------------------------------------------------------------------------------------------

def _bmax_unsubscribe_tool_tracker():
	if bmax_mode_owner is not None:
		bpy.msgbus.clear_by_owner(bmax_mode_owner)
	if bmax_tool_owner is not None:
		bpy.msgbus.clear_by_owner(bmax_tool_owner)
	if bmax_work_owner is not None:
		bpy.msgbus.clear_by_owner(bmax_work_owner)
	
#-----------------------------------------------------------------------------------------------------

@persistent
def bmax_tool_tracker_reload_handler(dummy):
	_bmax_subscribe_tool_tracker()

#-----------------------------------------------------------------------------------------------------
#
#						         BMax Tools registration
#
#-----------------------------------------------------------------------------------------------------

bmax_class_pool = (

	BMAX_OT_object_align,
	BMAX_OT_object_clone,
	BMAX_OT_object_mirror,
	BMAX_OT_object_unhide_all,
	BMAX_OT_object_hide_selected,
	BMAX_OT_object_hide_unselected,
	BMAX_OT_object_unfreeze_all,
	BMAX_OT_object_freeze_selected,
	BMAX_OT_object_freeze_unselected,

	BMAX_MT_object_align_special,
	BMAX_MT_object_hide,
	BMAX_MT_object_freeze,
	BMAX_MT_object_link,
	BMAX_MT_object_constraints,
	BMAX_MT_object_keyframing,

	BMAX_OT_mesh_make_planar_n,
	BMAX_OT_mesh_make_planar_x,
	BMAX_OT_mesh_make_planar_y,
	BMAX_OT_mesh_make_planar_z,
	BMAX_OT_mesh_view_align_x,
	BMAX_OT_mesh_view_align_y,
	BMAX_OT_mesh_view_align_z,
	BMAX_OT_mesh_sub_object_verts,
	BMAX_OT_mesh_sub_object_edges,
	BMAX_OT_mesh_sub_object_faces,
	BMAX_OT_mesh_sub_object_edge_verts,
	BMAX_OT_mesh_clone,
	BMAX_OT_mesh_turn_quad_diagonal,
	BMAX_OT_mesh_target_weld_on,
	BMAX_OT_mesh_target_weld_off,
	BMAX_OT_mesh_remove_isolated_geometry,

	BMAX_MT_mesh_align_special,
	BMAX_MT_mesh_sub_object,
	BMAX_MT_mesh_select_special,
	BMAX_MT_mesh_show_hide_element,
	BMAX_MT_mesh_clone,
	BMAX_MT_mesh_create,
	BMAX_MT_mesh_slice,
	BMAX_MT_mesh_meshtools,
	BMAX_MT_mesh_repair,

	BMAX_OT_armature_mirror,
	BMAX_OT_armature_extrude,
	BMAX_OT_armature_clone_bones,

	BMAX_MT_armature_create_primitive,
	BMAX_MT_armature_align_special,
	BMAX_MT_armature_bone_tools,
	BMAX_MT_armature_clone,
	BMAX_MT_armature_auto_names,
	BMAX_MT_armature_hide,

	BMAX_MT_pose_auto_names,
	BMAX_MT_pose_hide,
	BMAX_MT_pose_tools,
	BMAX_MT_pose_constraints,

	BMAX_OT_curve_clone,
	BMAX_MT_curve_hide,

	BMAX_MT_viewport,
	BMAX_OT_check_valid_faces,
	BMAX_PT_keyframing_settings,

	#-----------------------------------------

	BMAX_OT_uv_sub_object_verts,
	BMAX_OT_uv_sub_object_edges,
	BMAX_OT_uv_sub_object_faces,
	BMAX_OT_uv_sub_object_islands,
	BMAX_OT_uv_space_horizontal,
	BMAX_OT_uv_space_vertical,
	BMAX_OT_uv_rotate_plus,
	BMAX_OT_uv_rotate_minus,
	BMAX_OT_uv_flip_horizontal,
	BMAX_OT_uv_flip_vertical_uv,

	BMAX_MT_uv_sub_object,
	BMAX_MT_uv_show_hide_element,
	BMAX_MT_uv_tools,

	#-----------------------------------------

	BMAX_MT_graph_align_special,
	BMAX_MT_dopesheet_align_special,

	#-----------------------------------------

	BMAX_MT_tool_menu,
	BMAX_MT_tool_menu_uv,
	BMAX_MT_tool_menu_graph,
	BMAX_MT_tool_menu_dopesheet,

	#-----------------------------------------

	BMAX_OT_preview_default,
	BMAX_OT_preview_black_on_white,
	BMAX_OT_preview_white_on_black,
	BMAX_OT_preview_elsyiun,
	BMAX_OT_preview_softimage,
	BMAX_OT_preview_maya_light,
)

#-----------------------------------------------------------------------------------------------------
# setup
#-----------------------------------------------------------------------------------------------------

def register():
	for cls in bmax_class_pool:
		bpy.utils.register_class(cls)
	
	# setup shading extension
	bpy.types.VIEW3D_PT_shading_color.append(bmax_shading_extension)

	# setup tool tracker
	_bmax_subscribe_tool_tracker()

	if bmax_tool_tracker_reload_handler not in bpy.app.handlers.load_post:
		bpy.app.handlers.load_post.append(bmax_tool_tracker_reload_handler)
	
	# register keymap
	wm = bpy.context.window_manager

	km = wm.keyconfigs.addon.keymaps.new(name='Object Mode', space_type='EMPTY')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu"

	km = wm.keyconfigs.addon.keymaps.new(name='Mesh', space_type='EMPTY')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu"

	km = wm.keyconfigs.addon.keymaps.new(name='Armature', space_type='EMPTY')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu"

	km = wm.keyconfigs.addon.keymaps.new(name='Pose', space_type='EMPTY')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu"

	km = wm.keyconfigs.addon.keymaps.new(name='Curve', space_type='EMPTY')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu"

	km = wm.keyconfigs.addon.keymaps.new(name='UV Editor', space_type='EMPTY')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu_uv"

	km = wm.keyconfigs.addon.keymaps.new(name='Graph Editor', space_type='GRAPH_EDITOR')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu_graph"

	km = wm.keyconfigs.addon.keymaps.new(name='Dopesheet', space_type='DOPESHEET_EDITOR')
	kmi = km.keymap_items.new('wm.call_menu', 'RIGHTMOUSE', 'PRESS')
	kmi.properties.name = "BMAX_MT_tool_menu_dopesheet"

#-----------------------------------------------------------------------------------------------------

def unregister():
	for cls in reversed(bmax_class_pool):
		bpy.utils.unregister_class(cls)

	# remove shading extension
	bpy.types.VIEW3D_PT_shading_color.remove(bmax_shading_extension)

	# remove tool tracker
	_bmax_unsubscribe_tool_tracker()

	if bmax_tool_tracker_reload_handler in bpy.app.handlers.load_post:
		bpy.app.handlers.load_post.remove(bmax_tool_tracker_reload_handler)

	# remove keymap
	wm = bpy.context.window_manager

	km = wm.keyconfigs.addon.keymaps['Object Mode']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu":
				km.keymap_items.remove(kmi)

	km = wm.keyconfigs.addon.keymaps['Mesh']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu":
				km.keymap_items.remove(kmi)

	km = wm.keyconfigs.addon.keymaps['Armature']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu":
				km.keymap_items.remove(kmi)

	km = wm.keyconfigs.addon.keymaps['Pose']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu":
				km.keymap_items.remove(kmi)

	km = wm.keyconfigs.addon.keymaps['Curve']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu":
				km.keymap_items.remove(kmi)

	km = wm.keyconfigs.addon.keymaps['UV Editor']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu_uv":
				km.keymap_items.remove(kmi)

	km = wm.keyconfigs.addon.keymaps['Graph Editor']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu_graph":
				km.keymap_items.remove(kmi)

	km = wm.keyconfigs.addon.keymaps['Dopesheet']
	for kmi in km.keymap_items:
		if kmi.idname == 'wm.call_menu':
			if kmi.properties.name == "BMAX_MT_tool_menu_dopesheet":
				km.keymap_items.remove(kmi)

#-----------------------------------------------------------------------------------------------------
# entry point
#-----------------------------------------------------------------------------------------------------

if __name__ == "__main__":
	register()
